from collections import namedtuple
from typing import Callable, Collection, Iterable, List, Union
from tinytable.datatypes import RowDict, TableMapping
from tinytable.functional.filter import column_filter, filter_data

from tinytable.utils import combine_names_rows, row_dicts_to_data, uniques


TableFilter = Iterable[bool]


def groupbycolumn(data: TableMapping, column: Collection) -> List[tuple]:
    keys = uniques(column)
    return [(k, filter_data(data, column_filter(column, lambda x: x == k)))
                for k in keys]


def groupbyone(data: TableMapping, column_name: str) -> List[tuple]:
    return groupbycolumn(data, data[column_name])


def row_value_tuples(data: TableMapping, column_names: Collection[str]) -> List[tuple]:
    return list(zip(*[data[col] for col in column_names]))


def groupbymulti(data: TableMapping, column_names: Collection[str]) -> List[tuple]:
    return groupbycolumn(data, row_value_tuples(data, column_names))


def groupby(data: TableMapping, by: Union[str, Collection[str]]) -> List[tuple]:
    if isinstance(by, str):
        return groupbyone(data, by)
    else:
        return groupbymulti(data, by)


def _keys(key, by) -> dict:
    keys = {}
    if isinstance(by, str):
        keys[by] = key
    else:
        for col, k in zip(by, key):
            keys[col] = k
    return keys


def aggregate_groups(groups: List[tuple], by: Collection[str], func: Callable, tuplename: str) -> tuple[List, dict]:
    labels = []
    rows = []
    for key, data in groups:
        row = func(data)
        if len(row):
            GroupbyKey = namedtuple(field_names=by, typename='GroupbyKey')
            keys = _keys(key, by)
            labels.append(GroupbyKey(*keys.values()))
            rows.append(row)
    return labels, row_dicts_to_data(rows)


def sum_groups(groups: List[tuple], by: Collection[str]) -> tuple[List, dict]:
    return aggregate_groups(groups, by, sum_data, 'Sums')


def count_groups(groups: List[tuple], by: Collection[str]) -> tuple[List, dict]:
    return aggregate_groups(groups, by, count_data, 'Counts')


def aggregate_data(data: TableMapping, func: Callable) -> RowDict:
    out = {}
    for column_name in data.keys():
        try:
            col_sum = func(data[column_name])
        except TypeError:
            continue
        out[column_name] = col_sum
    return out


def sum_data(data: TableMapping) -> RowDict:
    return aggregate_data(data, sum)


def count_data(data: TableMapping) -> RowDict:
    return aggregate_data(data, len)





from typing import Optional, Union

import openpyxl
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.worksheet._read_only import ReadOnlyWorksheet
from openpyxl.chartsheet.chartsheet import Chartsheet
import tinytim as tim


Sheet = Union[Worksheet, ReadOnlyWorksheet, Chartsheet]
WorkSheet = Union[Worksheet, ReadOnlyWorksheet]


class WorkBook:
    def __init__(self, path: str) -> None:
        self.path = path

    def __enter__(self):
        self.wb = openpyxl.load_workbook(self.path)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.wb.close()

    @property
    def active(self) -> Worksheet:
        return self.wb.active

    def __getitem__(self, key: str) -> Sheet:
        return self.wb[key]


def read_excel_file(path: str, sheet_name: Optional[str] = None) -> dict:
    """
    Reads a table object from given excel file path.
    """
    column_names = []
    rows = []
    first = True
    with WorkBook(path) as wb:
        ws = wb.active if sheet_name is None else wb[sheet_name]
        if isinstance(ws, Chartsheet):
            raise TypeError('Chartsheet has no values to read into table.')
        for row in ws.values:
            if first:
                column_names = row
                first = False
            else:
                rows.append(row)

    return tim.utils.combine_names_rows(column_names, rows)


from __future__ import annotations
from typing import Any, Callable, MutableMapping, MutableSequence, Generator, Sequence, Union

from tabulate import tabulate
import tinytim as tim

from tinytable.filter import Filter
from tinytable.group import Group


class Column:
    def __init__(self, data: Sequence, name: Union[str, None], parent=None, labels=None):
        self.data = list(data)
        self.name = name
        self.parent = parent
        self.labels = labels

        
    def __len__(self) -> int:
        return len(self.data)
    
    def __repr__(self) -> str:
        header = 'index' if self.name is None else self.name
        index = True if self.labels is None else self.labels
        return tabulate({header: self.data}, headers=[header], tablefmt='grid', showindex=index)
    
    def __iter__(self):
        return iter(self.data)
    
    def __getitem__(self, index: int) -> Any:
        return self.data[index]
    
    def __setitem__(self, index: int, value: Any) -> None:
        self.data[index] = value
        if self.parent is not None:
            self.parent.edit_value(self.name, index, value)

    def __eq__(self, value: Any) -> Filter:
        return Filter(self, lambda x: x == value)

    def __ne__(self, value: Any) -> Filter:
        return Filter(self, lambda x: x != value)

    def __gt__(self, value: Any) -> Filter:
        return Filter(self, lambda x: x > value)

    def __lt__(self, value: Any) -> Filter:
        return Filter(self, lambda x: x < value)

    def __ge__(self, value: Any) -> Filter:
        return Filter(self, lambda x: x >= value)

    def __le__(self, value: Any) -> Filter:
        return Filter(self, lambda x: x <= value)

    def isin(self, values: MutableSequence) -> Filter:
        return Filter(self, lambda x: x in values)

    def notin(self, values: MutableSequence) -> Filter:
        return Filter(self, lambda x: x not in values)

    def drop(self):
        """drop Column from parent"""
        if self.parent is not None:
            self.parent.drop_column(self.name)
            self.parent = None

    def cast_as(self, data_type: Callable) -> None:
        self.data = [data_type(item) for item in self.data]
        if self.parent is not None:
            self.parent.cast_column_as(self.name, data_type)

    def value_counts(self) -> dict:
        return {value: self.data.count(value) for value in self.data}

    def sum(self) -> Union[float, int]:
        return sum(self.data)

    def groupby(self) -> Group:
        name = str(self.name)
        groups = tim.group.groupby({name: self.data}, by=str(name))
        return Group(groups, by=name)


def itercolumns(data: MutableMapping, parent, labels=None) -> Generator[Column, None, None]:
    for col in data.keys():
        yield Column(data[col], col, parent, labels)


def iteritems(data: MutableMapping, parent) -> Generator[tuple[str, Column], None, None]:
    for col in data.keys():
        yield col, Column(data[col], col, parent)


def cast_column_as(data: MutableMapping, column_name: str, data_type: Callable) -> MutableMapping:
    """Return a new dict with named column cast as data_type."""
    new_data = tim.copy.copy_table(data)
    new_data[column_name] = [data_type(value) for value in new_data[column_name]]
    return new_data


import sqlite3
from typing import List

import tinytim as tim


def _table_names(conn) -> List[str]:
    cursor = conn.cursor()
    cursor.execute('SELECT name from sqlite_master where type= "table"')
    return [x[0] for x in cursor.fetchall()]


def get_table_names(path: str) -> List[str]:
    with sqlite3.connect(path) as con:
        return _table_names(con)


def _table_column_names(conn, table_name):
    cursor = conn.execute(f'select * from {table_name}')
    return list(map(lambda x: x[0], cursor.description))


def _select_all_rows(conn, table_name):
    cur = conn.cursor()
    cur.execute(f"SELECT * FROM {table_name}")
    return cur.fetchall()


def read_sqlite_table(path: str, table_name: str) -> dict:
    with sqlite3.connect(path) as con:
        column_names = _table_column_names(con, table_name)
        rows = _select_all_rows(con, table_name)
    return tim.utils.combine_names_rows(column_names, rows)



import csv
from typing import Generator, Union

import tinytim as tim


def convert_str(value: str) -> Union[float, int, bool, str]:
    """Takes a str value and tries to convert it to float, int, or bool
       Returns converted value if successful, or str value if fails to convert.
    """
    value = str(value)
    if value.count('.') == 1:
        try:
            return float(value)
        except ValueError:
            pass
    if value.isnumeric():
        try:
            return int(value)
        except ValueError:
            pass
    if value in {'True', 'False'}:
        return bool(value)
    return value


def chunk_csv_file(
    path: str,
    chunksize=5,
    newline='',
    encoding='utf-8-sig'
) -> Generator[dict, None, None]:
    """
    Read chunks of table object from given CSV file.
    """
    column_names = []
    rows = []
    first = True
    chunk_end = chunksize
    with open(path, 'r', newline=newline, encoding=encoding) as f:
        dialect = csv.Sniffer().sniff(f.read(1024))
        f.seek(0)
        for i, row in enumerate(csv.reader(f, dialect)):
            if first:
                column_names = row
                first = False
            else:
                rows.append([convert_str(v) for v in row])
            if i == chunk_end:
                yield tim.utils.combine_names_rows(column_names, rows)
                rows = []
                chunk_end += chunksize
        else:
            if len(rows) > 0:
                yield tim.utils.combine_names_rows(column_names, rows)


def read_csv_file(
    path: str,
    newline='',
    encoding='utf-8-sig'
) -> dict:
    """
    Reads a table object from given CSV file path.
    """
    column_names = []
    rows = []
    first = True
    with open(path, 'r', newline=newline, encoding=encoding) as f:
        dialect = csv.Sniffer().sniff(f.read(1024))
        f.seek(0)
        for row in csv.reader(f, dialect):
            if first:
                column_names = row
                first = False
            else:
                rows.append([convert_str(v) for v in row])
        return tim.utils.combine_names_rows(column_names, rows)


from typing import Callable, Iterable, List, Collection, Sequence
from tinytable.datatypes import TableMapping


TableFilter = Iterable[bool]


def column_filter(column: Iterable, func: Callable) -> List[bool]:
    return [func(item) for item in column]


def indexes_from_filter(f: TableFilter) -> List[int]:
    return [i for i, b in enumerate(f) if b]


def filter_list_by_indexes(values: Sequence, indexes: Sequence[int]) -> List:
    """Return only values in indexes."""
    return [values[i] for i in indexes]


def filter_by_indexes(data: TableMapping, indexes: Sequence[int]) -> TableMapping:
    """Return only rows in indexes"""
    return {col: filter_list_by_indexes(values, indexes) for col, values in data.items()}


def filter_data(data: TableMapping, f: TableFilter) -> TableMapping:
    indexes = indexes_from_filter(f)
    return filter_by_indexes(data, indexes)


def filter_by_column_func(data: TableMapping, column_name: str, func) -> TableMapping:
    """Return only rows of data where named column equals value."""
    indexes = [i for i, val in enumerate(data[column_name]) if func(val)]
    return filter_by_indexes(data, indexes)


def filter_by_column_eq(data: TableMapping, column_name: str, value) -> TableMapping:
    """Return only rows of data where named column equals value."""
    return filter_by_column_func(data, column_name, lambda x: x == value)


def filter_by_column_ne(data: TableMapping, column_name: str, value) -> TableMapping:
    """Return only rows of data where named column does not equal value."""
    return filter_by_column_func(data, column_name, lambda x: x != value)


def filter_by_column_gt(data: TableMapping, column_name: str, value) -> TableMapping:
    """Return only rows of data where named column is greater than value."""
    return filter_by_column_func(data, column_name, lambda x: x > value)


def filter_by_column_lt(data: TableMapping, column_name: str, value) -> TableMapping:
    """Return only rows of data where named column is less than value."""
    return filter_by_column_func(data, column_name, lambda x: x < value)


def filter_by_column_ge(data: TableMapping, column_name: str, value) -> TableMapping:
    """Return only rows of data where named column is greater than or equal value."""
    return filter_by_column_func(data, column_name, lambda x: x >= value)


def filter_by_column_le(data: TableMapping, column_name: str, value) -> TableMapping:
    """Return only rows of data where named column is less than or equal value."""
    return filter_by_column_func(data, column_name, lambda x: x <= value)


def filter_by_column_isin(data: TableMapping, column_name: str, values) -> TableMapping:
    """Return only rows of data where named column is in values."""
    return filter_by_column_func(data, column_name, lambda x: x in values)


def filter_by_column_notin(data: TableMapping, column_name: str, values) -> TableMapping:
    """Return only rows of data where named column is not in values."""
    return filter_by_column_func(data, column_name, lambda x: x not in values)





from __future__ import annotations
import copy
from typing import Collection, Iterable, List, Mapping, MutableMapping, Optional, Union, Iterator
from typing import Any, Callable, MutableSequence, Generator

from tabulate import tabulate
import tinytim as tim

import tinytable.column as column
from tinytable.group import Group
import tinytable.row as row
from tinytable.csv import read_csv_file
from tinytable.excel import read_excel_file
from tinytable.sqlite import read_sqlite_table
from tinytable.filter import Filter


class Table:
    """Data table organized into {column_name: list[values]}
    
       A pure Python version of Pandas DataFrame.
    """
    def __init__(self, data: MutableMapping = {}, labels=None) -> None:
        self.data = data
        self._store_data()
        self._validate()
        self.labels = labels if labels is None else list(labels)

    def _store_data(self):
        for col in self.data:
            self._store_column(col, self.data[col])

    def _store_column(self, column_name: str, values: Iterable, inplace=True) -> Union[None, Table]:
        values = list(values)
        if inplace:
            tim.edit.edit_column(self.data, column_name, values)
        else:
            return Table(tim.edit.edit_column(self.data, column_name, values))
        
    def __len__(self) -> int:
        return tim.features.row_count(self.data)
        
    def __repr__(self) -> str:
        index = True if self.labels is None else self.labels
        return tabulate(self, headers=self.columns, tablefmt='grid', showindex=index)
    
    def __iter__(self) -> Iterator[str]:
        return iter(self.data)
    
    def __getitem__(self, key: Union[str, int]) -> Union[column.Column, row.Row, Table]:
        """
           Use str key for Column selection. Setting Column items changes parent Table.
           Use int key for Row selection. Setting Row items changes parent Table.

           Selecting subset of Table returns new Table,
           changes do not change original Table.
           Use int:int:int for index slice of Table rows.
           Use list of bool values to filter to Table of True rows.
           Use Filter to filter to Table of True rows.
        """
        # tbl['id'] -> Column
        if isinstance(key, str):
            return self.column(str(key))
        # tbl[1] -> Row
        if isinstance(key, int):
            index: int = self._convert_index(key)
            self._validate_index(index)
            return self.row(index)
        # tbl[1:4] -> Table
        if isinstance(key, slice):
            validate_int_slice(key)
            return self.filter_by_indexes(list(tim.utils.slice_to_range(key, len(self))))
        if isinstance(key, list):
            if tim.utils.all_bool(key):
                # tbl[[True, False, True, True]] -> Table
                return self.filter(key)
            # tbl[['id', 'name']] -> Table
            validate_list_key(key)
            return self.only_columns(key)
        if isinstance(key, Filter):
            # tbl[tbl['age'] >= 18] -> Table
            return self.filter(key)
        if isinstance(key, tuple):
            # tble[1, 2] or tbl[(1, 2)] -> labeled Row
            if self.labels is None:
                raise ValueError('Table must have labels to use tuple as key.')
            else:
                if key not in self.labels:
                    raise KeyError('tuple key is not in Table labels.')
                labels = self.labels
                if isinstance(labels, list):
                    index = labels.index(key)
                return self[index]
        raise TypeError('key must be str for column selection, int for row selection or slice for subset of Table rows.')
    
    def __setitem__(self, key: Union[str, int], values: MutableSequence) -> None:
        if type(key) == str:
            column_name: str = str(key)
            self.edit_column(column_name, values)
        elif type(key) == int:
            index: int = int(key)
            self.edit_row(index, values)

    def __delitem__(self, key: Union[str, int]) -> None:
        if type(key) == str:
            column_name: str = str(key)
            self.drop_column(column_name)
        elif type(key) == int:
            index: int = int(key)
            self.drop_row(index)

    @property
    def shape(self) -> tuple[int, int]:
        return tim.features.shape(self.data)

    @property
    def size(self) -> int:
        return tim.features.size(self.data)

    @property
    def columns(self) -> tuple[str]:
        """Column names."""
        return tim.features.column_names(self.data)

    @columns.setter
    def columns(self, values: MutableSequence) -> None:
        """Set the value of the column names."""
        self.replace_column_names(values)

    @property
    def index(self) -> column.Column:
        return column.Column(tim.features.index(self.data), None, self, self.labels)

    @property
    def values(self) -> tuple[tuple]:
        return tim.rows.values(self.data)

    def filter(self, f: Filter) -> Table:
        indexes = tim.filter.indexes_from_filter(f)
        return self.filter_by_indexes(indexes)

    def only_columns(self, column_names: MutableSequence[str]) -> Table:
        """Return new Table with only column_names Columns."""
        d = tim.filter.only_columns(self.data, column_names)
        return Table(d, self.labels)

    def _convert_index(self, index: int) -> int:
        if index < 0:
            return len(self) + index
        return index

    def _validate_index(self, index: int) -> None:
        if len(self) == 0:
            raise IndexError('row index out of range (empty Table)')
        upper_range = len(self) - 1
        if index > len(self) - 1 or index < 0:
            raise IndexError(f'row index {index} out of range (0-{upper_range})')
        
    def _validate(self) -> bool:
        count = None
        for key in self.data:
            col_count = len(self.data[key])
            if count is None:
                count = col_count
            if count != col_count:
                raise ValueError('All columns must be of the same length')
            count = col_count
        return True

    def _get_label(self, index: int) -> Union[None, List]:
        return None if self.labels is None else self.labels[index]
   
    def row(self, index: int) -> row.Row:
        label = self._get_label(index)
        return row.Row(tim.rows.row_dict(self.data, index), index, self, label)

    def column(self, column_name: str) -> column.Column:
        return column.Column(tim.features.column_values(self.data, column_name), column_name, self, self.labels)

    def drop_column(self, column_name: str, inplace=True) -> Union[None, Table]:
        if inplace:
            tim.edit.drop_column_inplace(self.data, column_name)
        else:
            return Table(tim.edit.drop_column(self.data, column_name), self.labels)

    def drop_row(self, index: int, inplace=True) -> Union[None, Table]:
        if inplace:
            tim.edit.drop_row_inplace(self.data, index)
            tim.edit.drop_label_inplace(self.labels, index)
        else:
            return Table(tim.edit.drop_row(self.data, index), tim.edit.drop_label(self.labels, index))

    def keys(self) -> tuple[str]:
        return self.columns
    
    def itercolumns(self) -> Generator[column.Column, None, None]:
        return column.itercolumns(self.data, self, self.labels)
            
    def iterrows(self) -> Generator[tuple[int, row.Row], None, None]:
        return row.iterrows(self.data, self, self.labels)

    def iteritems(self) -> Generator[tuple[str, column.Column], None, None]:
        return column.iteritems(self.data, self)

    def itertuples(self) -> Generator[tuple, None, None]:
        return tim.rows.itertuples(self.data)
    
    def edit_row(self, index: int, values: Union[Mapping, MutableSequence], inplace=True) -> Union[None, Table]:
        if inplace:
            if isinstance(values, Mapping):
                tim.edit.edit_row_items_inplace(self.data, index, values)
            elif isinstance(values, MutableSequence):
                tim.edit.edit_row_values_inplace(self.data, index, values)
        else:
            if isinstance(values, Mapping):
                data = tim.edit.edit_row_items(self.data, index, values)
            elif isinstance(values, MutableSequence):
                data = tim.edit.edit_row_values(self.data, index, values)
            return Table(data, copy.copy(self.labels))
            
    def edit_column(self, column_name: str, values: MutableSequence, inplace=True) ->Union[None, Table]:
        return self._store_column(column_name, values, inplace)

    def edit_value(self, column_name: str, index: int, value: Any, inplace=True) -> Union[None, Table]:
        if inplace:
            tim.edit.edit_value_inplace(self.data, column_name, index, value)
        else:
            return Table(tim.edit.edit_value(self.data, column_name, index, value), copy.copy(self.labels))

    def copy(self, deep=False) -> Table:
        if deep:
             return Table(tim.copy.deepcopy_table(self.data), copy.deepcopy(self.labels))
        return Table(tim.copy.copy_table(self.data), copy.copy(self.labels))

    def cast_column_as(self, column_name: str, data_type: Callable) -> None:
        self.data[column_name] = [data_type(value) for value in self.data[column_name]]

    def replace_column_names(self, new_keys: MutableSequence[str]) -> None:
        if len(new_keys) != len(self.data.keys()):
            raise ValueError('new_keys must be same len as dict keys.')
        for new_key, old_key in zip(new_keys, self.data.keys()):
            if new_key != old_key:
                self.data[new_key] = self.data[old_key]
                del self.data[old_key]

    def to_csv(self, path: str) -> None:
        """Save Table as csv at path."""
        ...

    def to_excel(self, path: str) -> None:
        """Save Table in Excel Workbook."""
        ...

    def to_sqlite(self, path: str, table_name: str) -> None:
        """Save Table in sqlite database."""
        ...

    def label_head(self, n: int = 5) -> Union[None, List]:
        return None if self.labels is None else self.labels[:5]

    def label_tail(self, n: int = 5) -> Union[None, List]:
        return None if self.labels is None else self.labels[5:]

    def head(self, n: int = 5) -> Table:
        return Table(tim.features.head(self.data, n), self.label_head(n))

    def tail(self, n: int = 5) -> Table:
        return Table(tim.features.tail(self.data, n), self.label_tail(n))

    def nunique(self) -> dict[str, int]:
        """Count number of distinct values in each column.
           Return dict with number of distinct values.
        """
        return tim.utils.nunique(self.data)

    def filter_by_indexes(self, indexes: MutableSequence[int]) -> Table:
        """return only rows in indexes"""
        labels = None if self.labels is None else tim.filter.filter_list_by_indexes(self.labels, indexes)
        return Table(tim.filter.filter_by_indexes(self.data, indexes), labels=labels)

    def sample(self, n, random_state=None) -> Table:
        """return random sample of rows"""
        indexes = tim.filter.sample_indexes(self.data, n, random_state)
        labels = None if self.labels is None else tim.filter.filter_list_by_indexes(self.labels, indexes)
        return Table(tim.filter.filter_by_indexes(self.data, indexes), labels=labels)

    def groupby(self, by: Union[str, Collection]) -> Group:
        return Group([(value, Table(data)) for value, data in tim.group.groupby(self.data, by)], by)

    def sum(self) -> dict:
        return tim.group.sum_data(self.data)

    def count(self) -> dict:
        return tim.group.count_data(self.data)


def read_csv(path: str):
    return Table(read_csv_file(path))


def read_excel(path: str, sheet_name: Optional[str] = None) -> Table:
    return Table(read_excel_file(path, sheet_name))


def read_sqlite(path: str, table_name: str) -> Table:
    return Table(read_sqlite_table(path, table_name))


def validate_int_slice(s: slice) -> None:
    if s.start is not None:
        if type(s.start) is not int:
            raise ValueError('slice start must be None or int')
    if s.stop is not None:
        if type(s.stop) is not int:
            raise ValueError('slice stop must be None or int')
    if s.step is not None:
        if type(s.step) is not int:
            raise ValueError('slice step must be None or int')


def validate_list_key(l: List) -> None:
    if not all(isinstance(item, str) for item in l):
        raise ValueError('All list items bust be str to use as key.')








import copy
import random
from typing import Any, List, MutableSequence, Generator, Optional, Mapping, Union

import tinytable.datatypes as dt
import tinytable.utils as utils
import tinytable.functional.inplace as inplace


def column_count(data: dt.TableMapping) -> int:
    """Return the number of columns in data."""
    return len(data)


def row_count(data: dt.TableMapping) -> int:
    """Return the number of rows in data."""
    if column_count(data) == 0: return 0
    return len(data[first_column_name(data)])


def shape(data: dt.TableMapping) -> tuple[int, int]:
    """Return data row count, column count tuple."""
    col_count = column_count(data)
    if col_count == 0: return 0, 0
    return row_count(data), col_count


def size(data: dt.TableMapping) -> int:
    """Return data row count multiplied by column count."""
    rows, columns = shape(data)
    return rows * columns


def first_column_name(data: dt.TableMapping) -> str:
    """Return the name of the first column.
       Raises StopIteration if data has zero columns.
    """
    return next(iter(data))


def column_names(data: dt.TableMapping) -> tuple[str]:
    """Return data column names."""
    return tuple(data)


def replace_column_names(data: dt.TableMapping, new_names: MutableSequence[str]) -> dt.TableDict:
    """Return a new dict same column data but new column names."""
    old_names = column_names(data)
    if len(new_names) != len(old_names):
        raise ValueError('new_names must be same size as data column_count.')
    return {new_name: data[old_name] for new_name, old_name in zip(new_names, old_names)}


def index(data: dt.TableMapping) -> tuple[int]:
    """Return tuple of data column indexes."""
    return tuple(range(row_count(data)))


def data_columns_same_len(data: dt.TableMapping) -> bool:
    """Check if data columns are all the same len."""
    if column_count(data) == 0: return True
    it = iter(data.values())
    the_len = len(next(it))
    return all(len(l) == the_len for l in it)


def valid_table_mapping(data: dt.TableMapping) -> bool:
    """Check if data is a true TableMapping."""
    if not utils.has_mapping_attrs(data): return False
    return data_columns_same_len(data)


def table_value(data: dt.TableMapping, column_name: str, index: int) -> Any:
    """Return one value from column at row index."""
    return data[column_name][index]


def row_dict(data: dt.TableMapping, index: int) -> dt.RowDict: 
    """Return one row from data at index."""
    return {col: table_value(data, col, index) for col in column_names(data)}


def row_values(data: dt.TableMapping, index: int) -> tuple:
    """Return a tuple of the values at row index."""
    return tuple(values[index] for values in data.values())


def column_dict(data, col: str) -> dict[str, MutableSequence]:
    return {col: data[col]}


def column_values(data: dt.TableMapping, column_name: str) -> MutableSequence:
    return data[column_name]


def itercolumns(data: dt.TableMapping) -> Generator[tuple[str, tuple], None, None]:
    """Return a generator of tuple column name, column values."""
    for col in column_names(data):
        yield col, tuple(data[col])
            

def iterrows(data: dt.TableMapping) -> Generator[tuple[int, dt.RowDict], None, None]:
    """Return a generator of tuple row index, row dict values."""
    for i in index(data):
        yield i, row_dict(data, i)


def itertuples(data: dt.TableMapping) -> Generator[tuple, None, None]:
    """Return a generator of tuple index and row values."""
    for i, row in iterrows(data):
        yield i, *row.values()


def itervalues(data: dt.TableMapping) -> Generator[tuple, None, None]:
    """Return a generator of tuple row values."""
    for i, row in iterrows(data):
        yield tuple(row.values())


def values(data: dt.TableMapping) -> tuple[tuple]:
    """Return tuple of tuple row values."""
    return tuple(itervalues(data))


def filter_by_indexes(data: dt.TableMapping, indexes: MutableSequence[int]) -> dt.TableDict:
    """return only rows in indexes"""
    return {col: [values[i] for i in indexes] for col, values in data.items()}


def only_columns(data: dt.TableMapping, column_names: MutableSequence[str]) -> dt.TableDict:
    """Return new TableDict with only column_names."""
    return {col: data[col] for col in column_names}


def sample(data: dt.TableMapping, n: int, random_state: Optional[int] = None) -> dt.TableDict:
    """return random sample of n rows"""
    if random_state is not None:
        random.seed(random_state)
    indexes = random.sample(range(row_count(data)), n)
    return filter_by_indexes(data, indexes)


def sample_indexes(data: dt.TableMapping, n: int, random_state: Optional[int] = None) -> List[int]:
    """return random sample of n indexes"""
    if random_state is not None:
        random.seed(random_state)
    return random.sample(range(row_count(data)), n)


def nunique(data: dt.TableMapping) -> dict[str, int]:
    """Count number of distinct values in each column.
       Return dict with number of distinct values.
    """
    return {col: len(utils.uniques(values)) for col, values in data.items()}


def head(data: dt.TableMapping, n: int = 5) -> dt.TableDict:
    """Return the first n rows of data."""
    return {col: values[:n] for col, values in data.items()}


def tail(data: dt.TableMapping, n: int = 5) -> dt.TableDict:
    """Return the last n rows of data."""
    return {col: values[-n:] for col, values in data.items()}


def edit_row_items(data: dt.TableMapping, index: int, items: Mapping) -> dt.TableMapping:
    """Return a new dict with row index changed to mapping items values."""
    new_data = copy.copy(data)
    inplace.edit_row_items(new_data, index, items)
    return new_data


def edit_row_values(data: dt.TableMapping, index: int, values: MutableSequence) -> dt.TableMapping:
    """Return a new dict with row index changed to values."""
    new_data = copy_table(data)
    inplace.edit_row_values(new_data, index, values)
    return new_data


def edit_column(data: dt.TableMapping, column_name: str, values: MutableSequence) -> dt.TableMapping:
    """Returns a new dict with values added to data in named column.
       Overrides existing values if column exists,
       Created new column with values if column does not exist.
    """
    new_data = copy_table(data)
    inplace.edit_column(data, column_name, values)
    return new_data


def edit_value(data: dt.TableMapping, column_name: str, index: int, value: Any) -> dt.TableMapping:
    """Return a new table with the value in named column changed at row index."""
    new_data = copy_table(data)
    inplace.edit_value(data, column_name, index, value)
    return new_data


def drop_row(data: dt.TableMapping, index: int) -> dt.TableMapping:
    """Return a new dict with index row removed from data."""
    new_data = copy_table(data)
    inplace.drop_row(data, index)
    return new_data


def drop_label(labels: Union[None, List], index: int) -> Union[None, List]:
    new_labels = copy.copy(labels)
    inplace.drop_label(new_labels, index)
    return new_labels


def drop_column(data: dt.TableMapping, column_name: str) -> dt.TableMapping:
    """Return a new dict with the named column removed from data."""
    new_data = copy_table(data)
    inplace.drop_column(data, column_name)
    return new_data


def copy_table(data: dt.TableMapping) -> dt.TableMapping:
    return copy.copy(data)


def deepcopy_table(data: dt.TableMapping) -> dt.TableMapping:
    return copy.deepcopy(data)


from __future__ import annotations
from typing import Callable


class FilterIterator:
    """Passed when iterating through Filter
       Applies filter func to parent Column values.
    """
    def __init__(self, filter: Filter) -> None:
        self.filter = filter
        self.iter = iter(self.filter.column)

    def __next__(self) -> bool:
        return self.filter.func(next(self.iter))

    def __iter__(self):
        return self


class Filter:
    """Object used to filter a Table by criteria.
       Returned when Column is used with boolean operator.
       Column == 1 or Column >= 10

       Pass as key in Table to filter to True rows.
       Table[Column > 1] -> Table where each row Column > 1
    """
    def __init__(self, column, func: Callable):
        self.column = column
        self.func = func

    def __iter__(self) -> FilterIterator:
        return FilterIterator(self)



from typing import Any, List, Mapping, MutableSequence, Union

import tinytable.datatypes as dt
import tinytable.functional.table as func


def edit_row_items(data: dt.TableMapping, index: int, items: Mapping) -> None:
    """Changes row index to mapping items values."""
    for col in items:
        data[col][index] = items[col]


def edit_row_values(data: dt.TableMapping, index: int, values: MutableSequence) -> None:
    """Changed row index to values."""
    if len(values) != func.column_count(data):
        raise AttributeError('values length must match columns length.')
    for col, value in zip(func.column_names(data), values):
        data[col][index] = value


def edit_column(data: dt.TableMapping, column_name: str, values: MutableSequence) -> None:
    """Add values to data in named column.
       Overrides existing values if column exists,
       Created new column with values if column does not exist.
    """
    if len(values) != func.row_count(data):
        raise ValueError('values length must match data rows count.')
    data[column_name] = values


def drop_row(data: dt.TableMapping, index: int) -> None:
    """Remove index row from data."""
    for col in func.column_names(data):
        data[col].pop(index)


def drop_label(labels: Union[None, List], index) -> None:
    if labels is not None:
        labels.pop(index)


def drop_column(data: dt.TableMapping, column_name: str) -> None:
    """Return a new dict with the named column removed from data."""
    del data[column_name]


def edit_value(data: dt.TableMapping, column_name: str, index: int, value: Any) -> None:
    """Edit the value in named column as row index."""
    data[column_name][index] = value



from typing import Any, Generator, List, MutableMapping, Dict, Tuple

from tabulate import tabulate
import tinytim as tim


class Row:
    def __init__(self, data: Dict[str, Any], index: int, parent=None, label=None):
        self.data = data
        self.index = index
        self.parent = parent
        self.label = label
        
    def __len__(self) -> int:
        return len(self.data)
    
    def __iter__(self):
        return tim.utils.row_values_generator(self.data)
    
    def __repr__(self) -> str:
        index = self.index if self.label is None else self.label
        return tabulate({col: [value] for col, value in self.data.items()}, headers=self.columns, tablefmt='grid', showindex=[index])
    
    def __getitem__(self, column: str) -> Any:
        return self.data[column]
    
    def __setitem__(self, column: str, value: Any) -> None:
        self.data[column] = value
        if self.parent is not None:
            self.parent.edit_value(column, self.index, value)
    
    @property
    def columns(self) -> List[str]:
        return list(self.data.keys())

    def keys(self) -> List[str]:
        return list(self.data.keys())

    def drop(self) -> None:
        """drop Row from parent"""
        if self.parent is not None:
            self.parent.drop_row(self.index)
            self.parent = None

    def values(self) -> List[Any]:
        return list(self.data.values())


def iterrows(data: MutableMapping, parent, labels=None) -> Generator[Tuple[int, Row], None, None]:
    if len(data) == 0:
        return
    i = 0
    while True:
        try:
            label = None if labels is None else labels[i]
            yield i, Row({col: data[col][i] for col in data}, i, parent, label)
        except IndexError:
            return
        i += 1





from collections import namedtuple
from typing import Collection, List, Union

import tinytim as tim

import tinytable as tt


class Group:
    """Returned by Column and Table groupby method.
       Acts like a list of tuple(key, Table)
       Can apply aggregation function to calculate new Table.
    """
    def __init__(self, groups: List[tuple], by: Union[str, Collection]):
        self.groups = groups
        self.by = by

    def __iter__(self):
        return iter(self.groups)

    def __repr__(self):
        return repr(self.groups)

    def __getitem__(self, i: int):
        return self.groups[i]
        
    def sum(self):
        labels, rows = tim.group.sum_groups(self.groups, self.by)
        return tt.Table(rows, labels)

    def count(self):
        labels, rows = tim.group.count_groups(self.groups, self.by)
        return tt.Table(rows, labels)

